// MSP430F148 demo1 -  using FreeRTOS ver 3.2.4
// Chris Barnart - ceb at dideas.com 
// http://support.dideas.com/index.php?title=FreeRTOS_for_MSP430F148
//
// based on MSP430F449 example programs from FreeRTOS 
// Version 1.00 - Feb 20, 2006
/////////////////////////////////////////////////////////////////////////////

/*
This program toggles some LEDs on bits 0-2 of PORT5
and has the following typcial output on UART1 (9600 baud)

MSP430F148 FreeRTOS demo program. (Feb 20 2006 17:58:44)

ZXYZZZZZZZZZYZZZZZZZZZZXYZZZZZZZZZZYZZZZZZZZZZXYZ
[989] ZZZZZZZZZYZZZZZZZZZZXYZZZZZZZZZZYZZZZZZZZZZXYZ
[986] ZZZZZZZZZYZZZZZZZZZZXYZZZZZZZZZZYZZZZZZZZZZXYZ
[986] ZZZZZZZZZYZZZZZZZZZZXYZZZZZZZZZZYZZZZZZZZZZXYZ
[986] ZZZZZZZZZYZZZZZZZZZZXYZZZZZZZZZZYZZZZZZZZZZXYZ
[985] ZZZZZZZZZYZZZZZZZZZZXYZZZZZZZZZZYZZZZZZZZZZXYZ
...
*/


/////////////////////////////////////////////////////////////////////////////
// defines for local platform

// My LED are on PORT5
#define P_LED_DIR (P5DIR)
#define P_LED_OUT (P5OUT)

// My LEDs are on BITs 0 to 2
#define LED_D1 (1<<0)
#define LED_D2 (1<<1)
#define LED_D3 (1<<2)
#define LED_ALL (LED_D1 | LED_D2| LED_D3)


// the file uart.c has some other constants that may need to be changed 
// by default UART1 is used on PORT3, bits 6 and 7
// select 'USE_UART0' to switch to UART0 on PORT3, bits 4,5

/////////////////////////////////////////////////////////////////////////////
// standard includes
#include <stdlib.h>
#include <signal.h>

#include <stdio.h>   // needed for printf

// FreeRTOS includes
#include "FreeRTOS.h"
#include "task.h"


// application includes
#include "loc_types.h"
#include "uart_isr.h"




/////////////////////////////////////////////////////////////////////////////

static void task_1( void *pvParameters );
static void task_2( void *pvParameters );
static void task_3( void *pvParameters );
static void task_idle( void *pvParameters );



/////////////////////////////////////////////////////////////////////////////
// the idle task run continiously - and samples xTaskGetTickCount()
// and increments a counter
// when ACCUMULATE_PERIOD_TICKS have elapsed, the counter value is output.
// A delay loop is used to calibrate the counter total to approximatly 1000
// when no other tasks are running.  
// Thus the free CPU cycles in percent times 10 is output....
// An output of 900 means 90.0% free.
//
// Unfortuanlly the loop value is dependent on the CPU clock freeqncy
// and will probably need to be calibrated.

#define ACCUMULATE_PERIOD_TICKS 1000
#define CALIBRATED_LOOP_CNT 781

void task_idle( void *pvParameters ) {
  unsigned int cnt, v,nv;
  int d;
  int i;


  v = xTaskGetTickCount();
  nv = v + ACCUMULATE_PERIOD_TICKS;
  cnt = 0;

  while(1) {

    while(1) {
      ++cnt;
      v = xTaskGetTickCount();
      d = nv-v;
      if (d <= 0) break;
      for(i=0;i < CALIBRATED_LOOP_CNT; i++) asm("nop");
    }
    
    printf("\n\r[%d] ",cnt);
    nv = v + ACCUMULATE_PERIOD_TICKS;
    cnt = 0;
  }
}



void task_1( void *pvParameters ) {

  portTickType xLastFlashTime = xTaskGetTickCount(); // get the current ticks 

 while(1) {
   P_LED_OUT ^= LED_D1;                  // toggle my LED - non atomic
   putchar('X');

// sleep this task for 500 ticks relative to last tick
   vTaskDelayUntil( &xLastFlashTime, 500 ); 
 }
}



void task_2( void *pvParameters ) {

 portTickType xLastFlashTime;
 xLastFlashTime = xTaskGetTickCount();

 while(1) {
   P_LED_OUT ^= LED_D2;
   putchar('Y');
   vTaskDelayUntil( &xLastFlashTime, 250 );
 }
}



void task_3( void *pvParameters ) {
 portTickType xLastFlashTime;
 xLastFlashTime = xTaskGetTickCount();

 while(1) {
   P_LED_OUT ^= LED_D3;
   putchar('Z');
   vTaskDelayUntil( &xLastFlashTime, 25 );
 }
}





int main( void ) {

  WDTCTL = WDTPW + WDTHOLD;    // Stop the watchdog

  // set MSP430 DSO to maximum freq - sec 4.3 - pg 4-15
  DCOCTL = DCO2 | DCO1 | DCO0;       
  BCSCTL1 = RSEL2 | RSEL1 | RSEL0;     

  // enable all the LEDS
  P_LED_DIR = LED_ALL;
  P_LED_OUT = LED_ALL;
  

  // setup UART1 to baud rate of 9600 and Tx, Rx queue of length 16 
  // clk_source is ACLK (SEL0), ACLK freq is 32768
  init_uart_isr(SSEL0, 9600, 32768, 16);   // setup the UART for the MSP430F148

  // output some stuff to serial port
  uart_putchar_isr_mode(0); // a little hack to use putchar without ISR
  printf("\n\nMSP430F148 FreeRTOS demo program. (%s %s)\n\n",__DATE__,__TIME__);
  uart_putchar_isr_mode(1); // return putchar to using the ISR

#if 1
#define WORKER_PRIORITY    (tskIDLE_PRIORITY + 3)  // add any small INT >= 1

  xTaskCreate(task_1, "", 50, NULL, WORKER_PRIORITY, NULL );
  xTaskCreate(task_2, "", 50, NULL, WORKER_PRIORITY, NULL );
  xTaskCreate(task_3, "", 50, NULL, WORKER_PRIORITY, NULL );
#endif

  xTaskCreate(task_idle, "", 150, NULL, tskIDLE_PRIORITY, NULL );	

  vTaskStartScheduler();       // start the FreeRTOS scheduler
	
  /* As the scheduler has been started the demo applications tasks will be
     executing and we should never get here! */
  return 0;
}
