// MSP430F148 UART support
// Chris Barnart - ceb at dideas.com :
// http://support.dideas.com/index.php?title=FreeRTOS_for_MSP430F148
// should be easy to select UART0 or 1 - for all MSP430 devices other than x12xx
//
// ISR portions of this code are from the FreeRTOS v 3.2.4 disto


// select a UART to use
// #define USE_UART_0 // UART0 is not fully tested
#define USE_UART_1




// Standard includes
#include <stdlib.h>
#include <signal.h>

// FreeRTOS scheduler includes
#include "FreeRTOS.h"
#include "queue.h"
#include "task.h"


// local includes
#include "loc_types.h"
#include "uart_isr.h"


#define DEFAULT_BLOCK_TICKS 100 // how long to block with putchar and getchar

#define serNO_BLOCK 0  // no block time value

#define ENABLE_AUTO_LF 1 // if set - then '\n' outputs '\r\n'


#ifdef USE_UART_0

// not valid for x12xx devices

#define UART_CTL           U0CTL
#define UART_ME            ME1
#define UART_ME_VAL        (UTXE0|URXE0)

#define UART_TX_BUF            U0TXBUF
#define UART_TX_CTRL       U0TCTL
#define UART_TX_EMP_TST    TXEPT


#define UART_RX_BUF            U0RXBUF
#define UART_RX_STATUS     IFG1         
#define UART_RX_RDY_TST    URXIFG0      

#define UART_IE            IE1          
#define UART_IE_TX         UTXIE0
#define UART_IE_RX         URXIE0

#define UART_BR0           U0BR0
#define UART_BR1           U0BR1
#define UART_MCTL          U0MCTL


#define UART_RX_ISR_VEC    (UART0RX_VECTOR)
#define UART_TX_ISR_VEC    (UART0TX_VECTOR)

#define UART_SEL_PINS      P3SEL                      // UART0 is on PORT3, bits B4 & B5
#define UART_SEL_BITS      (3 << 4)

#endif

#ifdef USE_UART_1

#define UART_CTL           U1CTL
#define UART_ME            ME2
#define UART_ME_VAL        (UTXE1|URXE1)

#define UART_TX_BUF            U1TXBUF
#define UART_TX_CTRL       U1TCTL
#define UART_TX_EMP_TST    TXEPT


#define UART_RX_BUF            U1RXBUF
#define UART_RX_STATUS     IFG2
#define UART_RX_RDY_TST    URXIFG1

#define UART_IE            IE2
#define UART_IE_TX         UTXIE1 
#define UART_IE_RX         URXIE1

#define UART_BR0           U1BR0
#define UART_BR1           U1BR1
#define UART_MCTL          U1MCTL

#define UART_RX_ISR_VEC    (UART1RX_VECTOR)
#define UART_TX_ISR_VEC    (UART1TX_VECTOR)

#define UART_SEL_PINS      P3SEL                      // UART1 is on PORT3, bits B6 & B7
#define UART_SEL_BITS      (3 << 6)
#endif



static xQueueHandle xRxedChars;  // queue for  received characters.
static xQueueHandle xCharsForTx; // queue for characters waiting transmission. 


static volatile portSHORT sTHREEmpty;

/* Interrupt service routines. */
interrupt (UART1RX_VECTOR) vRxISR( void );
interrupt (UART1TX_VECTOR) vTxISR( void );

char _putchar_isr_mode = 1; // putchar uses the ISR by default

// Compute the modular bit for the MSP430 uart
// Chris Barnhart - ceb@dideas.com - http://www.dideas.com
//
byte msp430_compute_modulator_bits(long baud, long freq, uint *p_div) {
  uint my_div;
  byte mod, j, m_tot;
  long p0, rem;

  my_div = freq/baud;
  rem = (baud * my_div) - freq;

  if (p_div) *p_div = my_div;

  m_tot = 0;
  mod = 0;
  p0 = 0;

  for(j=0;j<8;j++) {
    p0 += rem;
    if ((p0<<1) < -baud ) {
      ++m_tot;
      p0 += baud;
      mod |= (1<<j);
    } 
  } // for j

  return mod;
}


// Initialize the MSP430 uart and create FreeRTOS queues for the serial Tx and Rx ISR
//
// clk_source : UCLKI = 0, ACLK = SEL0, SMCLK = SEL1  - see sec 133, pg 13-23 (283) 
//
void init_uart_isr(byte clk_source, ulong baud, ulong uart_clk_freq, byte q_len) {
  uint my_div, mod;

  portENTER_CRITICAL();

  xRxedChars = xQueueCreate( q_len, (unsigned portBASE_TYPE) sizeof(signed portCHAR) );
  xCharsForTx = xQueueCreate( q_len, (unsigned portBASE_TYPE) sizeof(signed portCHAR) );

  UART_CTL = SWRST;
  UART_CTL |= CHAR;

  // SSEL0,1 control clock source :  UCLKI, ACLK, SMCLK, SMCLK
  UART_TX_CTRL = clk_source & (SSEL0 | SSEL1);

  mod = msp430_compute_modulator_bits(baud, uart_clk_freq, &my_div);

  UART_BR0 = my_div & 0xff;  
  UART_BR1 = (my_div >> 8) & 0xff;  
  UART_MCTL = mod;

  UART_ME |= UART_ME_VAL;

  UART_CTL = CHAR;

  UART_SEL_PINS |= UART_SEL_BITS;

  UART_IE |= UART_IE_TX | UART_IE_RX;

  portEXIT_CRITICAL();
}



int getchar(void) {
  int rcd;
  signed portCHAR ch;
  rcd = x_getchar(&ch, DEFAULT_BLOCK_TICKS);
  
  return (rcd) ? (int)ch : -1;
}


void uart_putchar_isr_mode(char val) {
  _putchar_isr_mode = val;
}


int putchar(int ch) {
  int rcd=0;

#ifdef ENABLE_AUTO_LF
  if (ch == '\n') putchar('\r');
#endif

  if (_putchar_isr_mode) {
    rcd = x_putchar(ch, DEFAULT_BLOCK_TICKS);
  }  else {
    while( !(UART_TX_CTRL & UART_TX_EMP_TST)) ;
    UART_TX_BUF = ch;
    rcd = 1;
  }
  return (rcd) ? ch : -1;
}




/////////////////////////////////////////////////////////////////////////////////////////
// The remainder of this code is taken from the demo/MSP430_GCC part of FreeRTOS v3.2.4

int x_getchar(signed portCHAR *pcRxedChar, portTickType xBlockTime ) {
  
  /* Get the next character from the buffer.  Return false if no characters
     are available, or arrive before xBlockTime expires. */

  if( xQueueReceive( xRxedChars, pcRxedChar, xBlockTime ) )	{
    return pdTRUE;
  }	else	{
    return pdFALSE;
  }
}



int x_putchar( signed portCHAR cOutChar, portTickType xBlockTime ) {
  signed portBASE_TYPE xReturn;

  /* Transmit a character. */

  portENTER_CRITICAL();
  {
    if( sTHREEmpty == pdTRUE )      {

      /* If sTHREEmpty is true then the UART Tx ISR has indicated that 
	 there are no characters queued to be transmitted - so we can
	 write the character directly to the shift Tx register. */
      sTHREEmpty = pdFALSE;
      
      UART_TX_BUF = cOutChar;

      xReturn = pdPASS;
    }    else      {
      /* sTHREEmpty is false, so there are still characters waiting to be
	 transmitted.  We have to queue this character so it gets 
	 transmitted	in turn. */
      
      /* Return false if after the block time there is no room on the Tx 
	 queue.  It is ok to block inside a critical section as each task
	 maintains it's own critical section status. */
      xReturn = xQueueSend( xCharsForTx, &cOutChar, xBlockTime );
      
      /* Depending on queue sizing and task prioritisation:  While we 
	 were blocked waiting to post on the queue interrupts were not 
	 disabled.  It is possible that the serial ISR has emptied the 
	 Tx queue, in which case we need to start the Tx off again
	 writing directly to the Tx register. */
      if( ( sTHREEmpty == pdTRUE ) && ( xReturn == pdPASS ) )	  {
	/* Get back the character we just posted. */
	xQueueReceive( xCharsForTx, &cOutChar, serNO_BLOCK );
	sTHREEmpty = pdFALSE;
	UART_TX_BUF = cOutChar;
      }
    }
  }
  portEXIT_CRITICAL();
  
  return pdPASS;
}




// UART RX interrupt service routine.

interrupt (UART_RX_ISR_VEC) vRxISR( void ) {
  signed portCHAR cChar;

  /* Get the character from the UART and post it on the queue of Rxed 
     characters. */
  cChar = UART_RX_BUF;
  
  if( xQueueSendFromISR( xRxedChars, &cChar, pdFALSE ) )     {
    /*If the post causes a task to wake force a context switch 
      as the woken task may have a higher priority than the task we have 
      interrupted. */
    taskYIELD();
  }
}





// UART Tx interrupt service routine.

interrupt (UART_TX_ISR_VEC) vTxISR( void ) {
  signed portCHAR cChar;
  portBASE_TYPE xTaskWoken;
  
  /* The previous character has been transmitted.  See if there are any
     further characters waiting transmission. */
  
  if( xQueueReceiveFromISR( xCharsForTx, &cChar, &xTaskWoken ) == pdTRUE )	{
    /* There was another character queued - transmit it now. */
    UART_TX_BUF = cChar;
  }	else	{
    /* There were no other characters to transmit. */
    sTHREEmpty = pdTRUE;
  }
}

